@@ -33,7 +33,6 @@
 #include <osmocom/core/socket.h>
 
 #include <osmocom/netif/stream.h>
-#include "xua_msg.h"
 
 #include <osmocom/sigtran/mtp_sap.h>
 #include <osmocom/sigtran/sccp_sap.h>
@@ -41,12 +40,14 @@
 #include <osmocom/sigtran/protocol/m3ua.h>
 #include <osmocom/sigtran/protocol/sua.h>
 
+#include "mtp3_hmdc.h"
 #include "xua_as_fsm.h"
 #include "xua_asp_fsm.h"
 #include "xua_internal.h"
 #include "ss7_as.h"
 #include "ss7_asp.h"
 #include "ss7_internal.h"
+#include "xua_msg.h"
 
 #define M3UA_MSGB_SIZE 1500
 
@@ -570,7 +571,7 @@ int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
  ***********************************************************************/
 
 /* obtain the destination point code from a M3UA message in XUA fmt * */
-struct m3ua_data_hdr *data_hdr_from_m3ua(struct xua_msg *xua)
+struct m3ua_data_hdr *data_hdr_from_m3ua(const struct xua_msg *xua)
 {
 	struct xua_msg_part *data_ie;
 	struct m3ua_data_hdr *data_hdr;
@@ -590,6 +591,7 @@ struct m3ua_data_hdr *data_hdr_from_m3ua(struct xua_msg *xua)
 /* This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
+	struct xua_msg_part *data_ie = xua_msg_find_tag(xua, M3UA_IEI_PROT_DATA);
 	struct xua_msg_part *na_ie = xua_msg_find_tag(xua, M3UA_IEI_NET_APPEAR);
 	struct xua_msg_part *rctx_ie = xua_msg_find_tag(xua, M3UA_IEI_ROUTE_CTX);
 	struct m3ua_data_hdr *dh;
@@ -621,23 +623,41 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		goto ret_free;
 	}
 
+	/* store the MTP-level information in the xua_msg for use by
+	 * higher layer protocols */
+	OSMO_ASSERT(data_ie);
+	dh = (struct m3ua_data_hdr *) data_ie->dat;
+	OSMO_ASSERT(dh);
+	m3ua_dh_to_xfer_param(&xua->mtp, dh);
+	LOGPASP(asp, DLM3UA, LOGL_DEBUG,
+		"%s(): M3UA data header: opc=%u=%s dpc=%u=%s sls=%u\n",
+		__func__, xua->mtp.opc, osmo_ss7_pointcode_print(asp->inst, xua->mtp.opc),
+		xua->mtp.dpc, osmo_ss7_pointcode_print2(asp->inst, xua->mtp.dpc),
+		xua->mtp.sls);
+
+	/* Drop packets not matching our configured Network Indicator: */
+	if (dh->ni != asp->inst->cfg.network_indicator) {
+		LOGPASP(asp, DLM3UA, LOGL_NOTICE,
+			"Discarding received XUA Message %s: NI=%u not matching ss7 instance configured NI=%u\n",
+			xua_hdr_dump(xua, &xua_dialect_sua), dh->ni, asp->inst->cfg.network_indicator);
+		rate_ctr_inc2(asp->inst->ctrg, SS7_INST_CTR_PKT_RX_NI_MISMATCH);
+		rate_ctr_inc2(asp->ctrg, SS7_ASP_CTR_PKT_RX_NI_MISMATCH);
+		rc = M3UA_ERR_UNEXPECTED_MSG;
+		goto ret_free;
+	}
+
 	rc = xua_find_as_for_asp(&as, asp, rctx_ie);
 	if (rc)
 		goto ret_free;
 
+	if (!osmo_ss7_as_active(as)) {
+		rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_DISCARD);
+		LOGPAS(as, DLM3UA, LOGL_INFO, "Discarding received XUA Message %s, AS state %s\n",
+		       xua_hdr_dump(xua, &xua_dialect_sua), osmo_fsm_inst_state_name(as->fi));
+		goto ret_free;
+	}
+
 	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
-
-	/* FIXME: check for AS state == ACTIVE */
-
-	/* store the MTP-level information in the xua_msg for use by
-	 * higher layer protocols */
-	dh = data_hdr_from_m3ua(xua);
-	OSMO_ASSERT(dh);
-	m3ua_dh_to_xfer_param(&xua->mtp, dh);
-	LOGPASP(asp, DLM3UA, LOGL_DEBUG,
-		"%s(): M3UA data header: opc=%u=%s dpc=%u=%s\n",
-		__func__, xua->mtp.opc, osmo_ss7_pointcode_print(asp->inst, xua->mtp.opc),
-		xua->mtp.dpc, osmo_ss7_pointcode_print2(asp->inst, xua->mtp.dpc));
 	OSMO_ASSERT(xua->mtp.sls <= 0xf);
 	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_SLS_0 + xua->mtp.sls);
 
@@ -648,7 +668,7 @@ static int m3ua_rx_xfer(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	}
 
 	/* xua ownership is passed here: */
-	return m3ua_hmdc_rx_from_l2(asp->inst, xua);
+	return mtp3_hmdc_rx_from_l2(asp->inst, xua);
 
 ret_free:
 	xua_msg_free(xua);
@@ -844,9 +864,9 @@ out:
  ***********************************************************************/
 
 /* 3.4.1 Destination Unavailable (DUNA) */
-static struct xua_msg *m3ua_encode_duna(const uint32_t *rctx, unsigned int num_rctx,
-					const uint32_t *aff_pc, unsigned int num_aff_pc,
-					const char *info_string)
+struct xua_msg *m3ua_encode_duna(const uint32_t *rctx, unsigned int num_rctx,
+				 const uint32_t *aff_pc, unsigned int num_aff_pc,
+				 const char *info_string)
 {
 	struct xua_msg *xua = xua_msg_alloc();
 
@@ -877,7 +897,6 @@ static struct xua_msg *m3ua_encode_dava(const uint32_t *rctx, unsigned int num_r
 	return xua;
 }
 
-#if 0 /* not used so far */
 /* 3.4.3 Destination Available (DAUD) */
 static struct xua_msg *m3ua_encode_daud(const uint32_t *rctx, unsigned int num_rctx,
 					const uint32_t *aff_pc, unsigned int num_aff_pc,
@@ -889,7 +908,6 @@ static struct xua_msg *m3ua_encode_daud(const uint32_t *rctx, unsigned int num_r
 		xua->hdr.msg_type = M3UA_SNM_DAUD;
 	return xua;
 }
-#endif
 
 /* 3.4.5 Destination User Part Unavailable (DUPU) */
 static struct xua_msg *m3ua_encode_dupu(const uint32_t *rctx, unsigned int num_rctx,
@@ -970,6 +988,21 @@ void m3ua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsi
 	m3ua_tx_xua_asp(asp, xua);
 }
 
+/*! Transmit SSNM DAUD message requesting [un]availability status of certain point code[s]
+ *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
+ *  \param[in] rctx array of Routing Contexts in network byte order.
+ *  \param[in] num_rctx number of rctx
+ *  \param[in] aff_pc array of 'Affected Point Code' in network byte order.
+ *  \param[in] num_aff_pc number of aff_pc
+ *  \param[in] info_string optional information string (can be NULL).
+ */
+void m3ua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+		      const uint32_t *aff_pc, unsigned int num_aff_pc, const char *info_string)
+{
+	struct xua_msg *xua = m3ua_encode_daud(rctx, num_rctx, aff_pc, num_aff_pc, info_string);
+	m3ua_tx_xua_asp(asp, xua);
+}
+
 /*! Transmit SSNM DUPU message indicating user unavailability.
  *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
  *  \param[in] rctx array of Routing Contexts in network byte order.
@@ -986,7 +1019,7 @@ void m3ua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int n
 }
 
 /* received SNM message on ASP side
- * This function takes ownership of xua msg passed to it. */
+ * xua is owned by parent call m3ua_rx_snm() */
 static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	struct osmo_ss7_as *as = NULL;
@@ -995,7 +1028,7 @@ static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 
 	rc = xua_find_as_for_asp(&as, asp, rctx_ie);
 	if (rc)
-		goto ret_free;
+		return rc;
 
 	/* report those up the stack so both other ASPs and local SCCP users can be notified */
 	switch (xua->hdr.msg_type) {
@@ -1032,28 +1065,47 @@ static int m3ua_rx_snm_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		rc = M3UA_ERR_UNSUPP_MSG_TYPE;
 	}
 
-ret_free:
-	xua_msg_free(xua);
 	return rc;
 }
 
-/* received SNM message on SG side */
+/* received SNM message on SG side
+ * xua is owned by parent call m3ua_rx_snm() */
 static int m3ua_rx_snm_sg(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
+	struct osmo_ss7_as *as = NULL;
+	struct xua_msg_part *rctx_ie;
+	int rc = 0;
+
 	switch (xua->hdr.msg_type) {
-	case M3UA_SNM_DAUD:	/* Audit: ASP inquires about availability of Point Codes */
+	case M3UA_SNM_SCON:
+		/* RFC4666 1.4.6: "The M3UA layer at an ASP or IPSP MAY indicate local congestion
+		 * to an M3UA peer with an SCON message."
+		 * RFC4666 3.4.4: The SCON message MAY also be sent from the M3UA layer of an ASP
+		 * to an M3UA peer, indicating that the congestion level of the M3UA layer or the
+		 * ASP has changed.*/
+		rctx_ie = xua_msg_find_tag(xua, M3UA_IEI_ROUTE_CTX);
+		rc = xua_find_as_for_asp(&as, asp, rctx_ie);
+		if (rc)
+			return rc;
+		xua_snm_rx_scon(asp, as, xua);
+		break;
+	case M3UA_SNM_DAUD:
+		/* Audit: ASP inquires about availability of Point Codes */
 		xua_snm_rx_daud(asp, xua);
 		break;
 	default:
 		return M3UA_ERR_UNSUPP_MSG_TYPE;
 	}
 
-	return 0;
+	return rc;
 }
 
+/* received SNM message
+ * This function takes ownership of xua msg passed to it. */
 static int m3ua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
 	struct xua_msg_part *na_ie = xua_msg_find_tag(xua, M3UA_IEI_NET_APPEAR);
+	int rc;
 
 	/* SNM only permitted in ACTIVE state */
 	if (asp->fi->state != XUA_ASP_S_ACTIVE) {
@@ -1064,7 +1116,8 @@ static int m3ua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		} else {
 			LOGPASP(asp, DLM3UA, LOGL_ERROR, "Rx M3UA SNM not permitted "
 				"while ASP in state %s\n", osmo_fsm_inst_state_name(asp->fi));
-			return M3UA_ERR_UNEXPECTED_MSG;
+			rc = M3UA_ERR_UNEXPECTED_MSG;
+			goto ret_free;
 		}
 	}
 
@@ -1076,16 +1129,25 @@ static int m3ua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 			"Unsupported 'Network Appearance' IE '0x%08x' in message type '%s', sending 'Error'.\n",
 			na, get_value_string(m3ua_xfer_msgt_names, xua->hdr.msg_type));
 		if (na_ie->len != 4)
-			return M3UA_ERR_PARAM_FIELD_ERR;
-		return M3UA_ERR_INVAL_NET_APPEAR;
+			rc = M3UA_ERR_PARAM_FIELD_ERR;
+		else
+			rc = M3UA_ERR_INVAL_NET_APPEAR;
+		goto ret_free;
 	}
 
 	switch (asp->cfg.role) {
 	case OSMO_SS7_ASP_ROLE_SG:
-		return m3ua_rx_snm_sg(asp, xua);
+		rc = m3ua_rx_snm_sg(asp, xua);
+		break;
 	case OSMO_SS7_ASP_ROLE_ASP:
-		return m3ua_rx_snm_asp(asp, xua);
+		rc = m3ua_rx_snm_asp(asp, xua);
+		break;
 	default:
-		return M3UA_ERR_UNSUPP_MSG_CLASS;
+		rc = M3UA_ERR_UNSUPP_MSG_CLASS;
+		break;
 	}
+
+ret_free:
+	xua_msg_free(xua);
+	return rc;
 }
