@@ -15,8 +15,7 @@ extern bool ss7_initialized;
 
 bool ss7_ipv6_sctp_supported(const char *host, bool bind);
 
-struct osmo_ss7_as *ss7_as_alloc(struct osmo_ss7_instance *inst, const char *name,
-				 enum osmo_ss7_asp_protocol proto);
+uint32_t ss7_pc_full_mask(const struct osmo_ss7_pc_fmt *pc_fmt);
 
 struct osmo_ss7_asp *ss7_asp_find_by_socket_addr(int fd, int trans_proto);
 
@@ -28,3 +27,6 @@ int ss7_asp_m3ua_tcp_srv_conn_rx_cb(struct osmo_stream_srv *conn, int res, struc
 int ss7_asp_xua_srv_conn_closed_cb(struct osmo_stream_srv *srv);
 
 int xua_tcp_segmentation_cb(struct msgb *msg);
+
+/* VTY */
+#define XUA_VAR_STR	"(sua|m3ua|ipa)"
