@@ -34,18 +34,21 @@
 #include <osmocom/core/fsm.h>
 
 #include <osmocom/netif/stream.h>
-#include <osmocom/sigtran/xua_msg.h>
+#include "xua_msg.h"
 
 #include <osmocom/sigtran/sccp_sap.h>
 #include <osmocom/sigtran/protocol/mtp.h>
 #include <osmocom/sigtran/protocol/sua.h>
 #include <osmocom/sigtran/protocol/m3ua.h>
 #include <osmocom/sigtran/osmo_ss7.h>
+#include <osmocom/sccp/sccp_types.h>
 
 #include "xua_as_fsm.h"
 #include "xua_asp_fsm.h"
 #include "xua_internal.h"
 #include "sccp_internal.h"
+#include "ss7_as.h"
+#include "ss7_asp.h"
 #include "ss7_internal.h"
 
 /* Appendix C.4 of Q.714 (all in milliseconds) */
@@ -89,7 +92,7 @@ static const struct value_string sua_iei_names[] = {
 	{ SUA_IEI_DEST_ADDR,		"Destination Address" },
 	{ SUA_IEI_SRC_REF,		"Source Reference" },
 	{ SUA_IEI_DEST_REF,		"Destination Reference" },
-	{ SUA_IEI_CAUSE,		"Cause" },
+	{ SUA_IEI_CAUSE,		"SCCP Cause" },
 	{ SUA_IEI_SEQ_NR,		"Sequence Number" },
 	{ SUA_IEI_RX_SEQ_NR,		"Receive Sequence Number" },
 	{ SUA_IEI_ASP_CAPA,		"ASP Capability" },
@@ -259,6 +262,54 @@ static struct xua_msg *sua_gen_error_msg(uint32_t err_code, struct msgb *msg)
 	return xua;
 }
 
+/* generate a return/refusal message (SUA CLDR == SCCP UDTS) based on
+ * the incoming message.  We need to flip all identities between sender
+ * and receiver */
+struct xua_msg *sua_gen_cldr(const struct xua_msg *xua_in, uint32_t route_ctx, uint32_t ret_cause)
+{
+	struct xua_msg *xua_out = xua_msg_alloc();
+	struct osmo_sccp_addr called;
+
+	xua_out->hdr = XUA_HDR(SUA_MSGC_CL, SUA_CL_CLDR);
+	xua_msg_add_u32(xua_out, SUA_IEI_ROUTE_CTX, route_ctx);
+	xua_msg_add_u32(xua_out, SUA_IEI_CAUSE,
+			SUA_CAUSE_T_RETURN | ret_cause);
+	/* Swap Calling and Called Party */
+	xua_msg_copy_part(xua_out, SUA_IEI_SRC_ADDR, xua_in, SUA_IEI_DEST_ADDR);
+	xua_msg_copy_part(xua_out, SUA_IEI_DEST_ADDR, xua_in, SUA_IEI_SRC_ADDR);
+	/* TODO: Optional: Hop Count */
+	/* Optional: Importance */
+	xua_msg_copy_part(xua_out, SUA_IEI_IMPORTANCE,
+			  xua_in, SUA_IEI_IMPORTANCE);
+	/* Optional: Message Priority */
+	xua_msg_copy_part(xua_out, SUA_IEI_MSG_PRIO, xua_in, SUA_IEI_MSG_PRIO);
+	/* Optional: Correlation ID */
+	xua_msg_copy_part(xua_out, SUA_IEI_CORR_ID, xua_in, SUA_IEI_CORR_ID);
+	/* Optional: Segmentation */
+	xua_msg_copy_part(xua_out, SUA_IEI_SEGMENTATION,
+			  xua_in, SUA_IEI_SEGMENTATION);
+	/* Optional: Data */
+	xua_msg_copy_part(xua_out, SUA_IEI_DATA, xua_in, SUA_IEI_DATA);
+
+	OSMO_ASSERT(sua_addr_parse(&called, xua_out, SUA_IEI_DEST_ADDR) == 0);
+
+	/* Route on PC + SSN ? */
+	if (called.ri == OSMO_SCCP_RI_SSN_PC) {
+		/* if no PC, copy OPC into called addr */
+		if (!(called.presence & OSMO_SCCP_ADDR_T_PC)) {
+			struct osmo_sccp_addr calling;
+			OSMO_ASSERT(sua_addr_parse(&calling, xua_out, SUA_IEI_SRC_ADDR) == 0);
+			called.presence |= OSMO_SCCP_ADDR_T_PC;
+			called.pc = calling.pc;
+			/* Re-encode / replace called address */
+			xua_msg_free_tag(xua_out, SUA_IEI_DEST_ADDR);
+			xua_msg_add_sccp_addr(xua_out, SUA_IEI_DEST_ADDR,
+					      &called);
+		}
+	}
+	return xua_out;
+}
+
 /***********************************************************************
  * Transmitting SUA messages to SCTP
  ***********************************************************************/
@@ -288,13 +339,16 @@ static struct msgb *sua_to_msg(struct xua_msg *xua)
 
 static int sua_tx_xua_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 {
-	struct msgb *msg = sua_to_msg(xua);
+	struct msgb *msg;
 
 	OSMO_ASSERT(asp->cfg.proto == OSMO_SS7_ASP_PROT_SUA);
 
+	msg = sua_to_msg(xua);
+	xua_msg_free(xua);
 	if (!msg)
 		return -1;
 
+	/* msg becomes owned by osmo_ss7_asp_send here: */
 	return osmo_ss7_asp_send(asp, msg);
 }
 
@@ -304,7 +358,6 @@ static int sua_tx_xua_asp(struct osmo_ss7_asp *asp, struct xua_msg *xua)
  *  \return 0 on success; negative on error */
 int sua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 {
-	struct msgb *msg;
 	int rc;
 
 	OSMO_ASSERT(as->cfg.proto == OSMO_SS7_ASP_PROT_SUA);
@@ -313,15 +366,11 @@ int sua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua)
 	if (as->cfg.routing_key.context)
 		xua_msg_add_u32(xua, SUA_IEI_ROUTE_CTX, as->cfg.routing_key.context);
 
-	msg = sua_to_msg(xua);
-	if (!msg)
-		return -1;
-
 	/* send the msg to the AS for transmission.  The AS FSM might
 	 * (depending on its state) enqueue it before transmission */
-	rc = osmo_fsm_inst_dispatch(as->fi, XUA_AS_E_TRANSFER_REQ, msg);
+	rc = osmo_fsm_inst_dispatch(as->fi, XUA_AS_E_TRANSFER_REQ, xua);
 	if (rc < 0)
-		msgb_free(msg);
+		xua_msg_free(xua);
 	return rc;
 }
 
@@ -529,6 +578,8 @@ static int sua_rx_cl(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		return rc;
 
 	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
+	OSMO_ASSERT(xua->mtp.sls <= 0xf);
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_SLS_0 + xua->mtp.sls);
 
 	switch (xua->hdr.msg_type) {
 	case 0: /* Reserved, permitted by ETSI TS 101 592 5.2.3.2 */
@@ -559,6 +610,8 @@ static int sua_rx_co(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		return rc;
 
 	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_TOTAL);
+	OSMO_ASSERT(xua->mtp.sls <= 0xf);
+	rate_ctr_inc2(as->ctrg, SS7_AS_CTR_RX_MSU_SLS_0 + xua->mtp.sls);
 
 	switch (xua->hdr.msg_type) {
 	case 0: /* Reserved, permitted by ETSI TS 101 592 5.2.3.2 */
@@ -712,7 +765,7 @@ int sua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg)
 	xua->mtp.opc = ;
 	xua->mtp.dpc = ;
 #endif
-	xua->mtp.sio = MTP_SI_SCCP;
+	xua->mtp.sio = MTP_SIO(MTP_SI_SCCP, asp->inst->cfg.network_indicator);
 
 	LOGPASP(asp, DLSUA, LOGL_DEBUG, "Received SUA Message (%s)\n",
 		xua_hdr_dump(xua, &xua_dialect_sua));
@@ -820,7 +873,6 @@ static struct xua_msg *sua_encode_dava(const uint32_t *rctx, unsigned int num_rc
 	return xua;
 }
 
-#if 0 /* not used so far */
 /* 3.4.3 Destination Available (DAUD) */
 static struct xua_msg *sua_encode_daud(const uint32_t *rctx, unsigned int num_rctx,
 					const uint32_t *aff_pc, unsigned int num_aff_pc,
@@ -832,7 +884,6 @@ static struct xua_msg *sua_encode_daud(const uint32_t *rctx, unsigned int num_rc
 		xua->hdr.msg_type = SUA_SNM_DAUD;
 	return xua;
 }
-#endif
 
 /* 3.4.5 Destination User Part Unavailable (DUPU) */
 static struct xua_msg *sua_encode_dupu(const uint32_t *rctx, unsigned int num_rctx,
@@ -858,7 +909,7 @@ static struct xua_msg *sua_encode_dupu(const uint32_t *rctx, unsigned int num_rc
 }
 
 /*! Transmit SSNM DUNA/DAVA message indicating [un]availability of certain point code[s]
- *  \param[in] asp ASP through whihc to transmit message. Must be ACTIVE.
+ *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
  *  \param[in] rctx array of Routing Contexts in network byte order.
  *  \param[in] num_rctx number of rctx
  *  \param[in] aff_pc array of 'Affected Point Code' in network byte order.
@@ -909,6 +960,23 @@ void sua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsig
 		xua_msg_add_data(xua, SUA_IEI_INFO_STRING, strlen(info_string)+1, (const uint8_t *) info_string);
 }
 
+/*! Transmit SSNM DAUD message requesting [un]availability status of certain point code[s]
+ *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
+ *  \param[in] rctx array of Routing Contexts in network byte order.
+ *  \param[in] num_rctx number of rctx
+ *  \param[in] aff_pc array of 'Affected Point Code' in network byte order.
+ *  \param[in] num_aff_pc number of aff_pc
+ *  \param[in] aff_ssn affected SSN (optional)
+ *  \param[in] smi subsystem multiplicity indicator (optional)
+ *  \param[in] info_string optional information string (can be NULL).
+ */
+void sua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+		     const uint32_t *aff_pc, unsigned int num_aff_pc, const uint32_t *aff_ssn,
+		     const uint32_t *smi, const char *info_string)
+{
+	struct xua_msg *xua = sua_encode_daud(rctx, num_rctx, aff_pc, num_aff_pc, aff_ssn, smi, info_string);
+	sua_tx_xua_asp(asp, xua);
+}
 
 /*! Transmit SSNM DUPU message indicating user unavailability.
  *  \param[in] asp ASP through which to transmit message. Must be ACTIVE.
@@ -1008,3 +1076,53 @@ static int sua_rx_snm(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		return SUA_ERR_UNSUPP_MSG_CLASS;
 	}
 }
+
+/* RFC 3868 3.10.6. SCCP Cause */
+const struct value_string sua_cause_type_names[] = {
+	{ SUA_CAUSE_T_RETURN, "Return Cause" },
+	{ SUA_CAUSE_T_REFUSAL, "Refusal Cause" },
+	{ SUA_CAUSE_T_RELEASE, "Release Cause" },
+	{ SUA_CAUSE_T_RESET, "Reset Cause" },
+	{ SUA_CAUSE_T_ERROR, "Error Cause" },
+	{}
+};
+const char *osmo_sua_sccp_cause_name(uint32_t sccp_cause, char *out_buf, size_t out_buf_len)
+{
+	uint16_t cause_type = sccp_cause & SUA_CAUSE_T_MASK;
+	uint8_t cause_val = sccp_cause & 0xff;
+	const char *cause_type_str = get_value_string_or_null(sua_cause_type_names, cause_type);
+	const char *cause_str = NULL;
+
+	if (!out_buf || out_buf_len == 0)
+		return NULL;
+
+	if (!cause_type_str)
+		goto ret_unknown_cause_type;
+
+	switch (cause_type) {
+	case SUA_CAUSE_T_RETURN:
+		cause_str = osmo_sccp_return_cause_name(cause_val);
+		break;
+	case SUA_CAUSE_T_REFUSAL:
+		cause_str = osmo_sccp_refusal_cause_name(cause_val);
+		break;
+	case SUA_CAUSE_T_RELEASE:
+		cause_str = osmo_sccp_release_cause_name(cause_val);
+		break;
+	case SUA_CAUSE_T_RESET:
+		cause_str = osmo_sccp_reset_cause_name(cause_val);
+		break;
+	case SUA_CAUSE_T_ERROR:
+		cause_str = osmo_sccp_error_cause_name(cause_val);
+		break;
+	default:
+		goto ret_unknown_cause_type;
+	}
+
+	snprintf(out_buf, out_buf_len, "%s <%s>", cause_type_str, cause_str);
+	return out_buf;
+
+ret_unknown_cause_type:
+	snprintf(out_buf, out_buf_len, "Unknown SUA SCCP Cause 0x%08x\n", sccp_cause);
+	return out_buf;
+}
