@@ -9,6 +9,8 @@
  * osmo-ss7.git
  */
 
+#include "config.h"
+
 #include <errno.h>
 
 #include <osmocom/core/fsm.h>
@@ -31,10 +33,15 @@
 #include "ss7_asp.h"
 #include "ss7_internal.h"
 #include "ss7_xua_srv.h"
+#include "sccp_internal.h"
 #include "xua_asp_fsm.h"
 #include "xua_as_fsm.h"
 #include "xua_internal.h"
 
+#ifdef WITH_TCAP_LOADSHARING
+#include "tcap_as_loadshare.h"
+#endif /* WITH_TCAP_LOADSHARING */
+
 #define S(x)	(1 << (x))
 
 /* The general idea is:
@@ -44,9 +51,6 @@
  * *
  */
 
-/* According to RFC3868 Section 8 */
-#define XUA_T_ACK_SEC	2
-
 static const struct value_string xua_asp_event_names[] = {
 	{ XUA_ASP_E_M_ASP_UP_REQ,	"M-ASP_UP.req" },
 	{ XUA_ASP_E_M_ASP_ACTIVE_REQ,	"M-ASP_ACTIVE.req" },
@@ -88,10 +92,16 @@ struct xua_asp_fsm_priv {
 	/* routing context[s]: list of 32bit integers */
 	/* ACTIVE: traffic mode type, tid label, drn label ? */
 
-	struct {
+	struct { /* RFC3868 & RFC4666 timer T(ack) */
 		struct osmo_timer_list timer;
 		int out_event;
 	} t_ack;
+
+	/* Timer for tracking HEARTBEAT without HEARTBEAT ACK */
+	struct {
+		struct osmo_timer_list timer;
+		uint32_t unacked_beats;
+	} t_beat;
 };
 
 struct osmo_xlm_prim *xua_xlm_prim_alloc(enum osmo_xlm_prim_type prim_type,
@@ -143,62 +153,65 @@ static void send_xlm_prim_simple(struct osmo_fsm_inst *fi,
 	xua_asp_send_xlm_prim_simple(asp, prim_type, op);
 }
 
-/* determine the osmo_ss7_as_traffic_mode to be used by this ASP; will
- * iterate over all AS configured for this ASP.  If they're compatible,
- * a single traffic mode is returned as enum osmo_ss7_as_traffic_mode.
- * If they're incompatible, -EINVAL is returned.  If there is none
- * configured, -1 is returned */
-static int determine_traf_mode(struct osmo_ss7_asp *asp)
+static void xua_asp_tx_snm_daud_address_book(struct osmo_ss7_asp *asp)
 {
-	struct osmo_ss7_as *as;
-	int tmode = -1;
+	struct osmo_ss7_instance *inst = asp->inst;
+	uint32_t rctx[OSMO_SS7_MAX_RCTX_COUNT];
+	unsigned int num_rctx;
+	uint32_t *aff_pc = NULL;
+	unsigned int num_aff_pc = 0;
+	struct osmo_sccp_addr_entry *entry;
 
-	llist_for_each_entry(as, &asp->inst->as_list, list) {
-		if (!osmo_ss7_as_has_asp(as, asp))
+	num_rctx = ss7_asp_get_all_rctx_be(asp, rctx, ARRAY_SIZE(rctx), NULL);
+
+	/* First count required size of num_aff_pc array: */
+	llist_for_each_entry(entry, &inst->cfg.sccp_address_book, list) {
+		if (!(entry->addr.presence & OSMO_SCCP_ADDR_T_PC))
 			continue;
-		/* we only care about traffic modes explicitly set */
-		if (!as->cfg.mode_set_by_vty)
+		if (osmo_ss7_pc_is_local(inst, entry->addr.pc))
 			continue;
-		if (tmode == -1) {
-			/* this is the first AS; we use this traffic mode */
-			tmode = as->cfg.mode;
-		} else {
-			if (tmode != as->cfg.mode)
-				return -EINVAL;
-		}
+		num_aff_pc++;
 	}
-	return tmode;
+	if (num_aff_pc == 0) {
+		LOGPASP(asp, DLSS7, LOGL_NOTICE, "Skip Tx DAUD: No SCCP in address book\n");
+		return;
+	}
+	aff_pc = talloc_array(asp, uint32_t, num_aff_pc);
+	OSMO_ASSERT(aff_pc);
+
+	num_aff_pc = 0;
+	llist_for_each_entry(entry, &inst->cfg.sccp_address_book, list) {
+		uint32_t curr_aff_pc;
+		unsigned int i;
+		if (!(entry->addr.presence & OSMO_SCCP_ADDR_T_PC))
+			continue;
+		if (osmo_ss7_pc_is_local(inst, entry->addr.pc))
+			continue;
+		LOGPASP(asp, DLSS7, LOGL_DEBUG, "Tx DAUD: Requesting status of DPC=%u=%s\n",
+			entry->addr.pc, osmo_ss7_pointcode_print2(inst, entry->addr.pc));
+		curr_aff_pc = htonl(entry->addr.pc); /* mask = 0 */
+		for (i = 0; i < num_aff_pc; i++)
+			if (aff_pc[i] == curr_aff_pc)
+				break;
+		if (i == num_aff_pc) /* not found in array */
+			aff_pc[num_aff_pc++] = curr_aff_pc;
+	}
+
+	xua_tx_snm_daud(asp, rctx, num_rctx, aff_pc, num_aff_pc, "Isolation-ASP-ACTIVE");
+	talloc_free(aff_pc);
 }
 
 /* add M3UA_IEI_ROUTE_CTX to xua_msg containig all routing keys of ASs within ASP */
 static int xua_msg_add_asp_rctx(struct xua_msg *xua, struct osmo_ss7_asp *asp)
 {
-	struct osmo_ss7_as *as;
-	uint32_t rctx[128];
-	unsigned int i = 0;
-
-	/* iterate over all ASs and build array of routing contexts */
-	llist_for_each_entry(as, &asp->inst->as_list, list) {
-		if (!osmo_ss7_as_has_asp(as, asp))
-			continue;
-		rctx[i++] = htonl(as->cfg.routing_key.context);
-		if (i >= ARRAY_SIZE(rctx)-1) {
-			break;
-		}
-	}
-	/* add xUA IE with routing contests to the message (if any) */
-	if (i) {
-		/* bail out (and not add the IE) if there's only one routing context (and hence
-		 * only one AS) within this ASP, and that routing context is zero, meaning no routing
-		 * context IE shall be used */
-		if (i == 1 && rctx[0] == 0)
-			return 0;
-
-		xua_msg_add_data(xua, M3UA_IEI_ROUTE_CTX, i*sizeof(uint32_t), (uint8_t *)rctx);
-	}
+	uint32_t rctx[OSMO_SS7_MAX_RCTX_COUNT];
+	unsigned int cnt;
 
+	cnt = ss7_asp_get_all_rctx_be(asp, rctx, ARRAY_SIZE(rctx), NULL);
+	if (cnt > 0)
+		xua_msg_add_data(xua, M3UA_IEI_ROUTE_CTX, cnt*sizeof(uint32_t), (uint8_t *)rctx);
 	/* return count of routing contexts added */
-	return i;
+	return cnt;
 }
 
 /* ask the xUA implementation to transmit a specific message */
@@ -215,13 +228,33 @@ static int peer_send(struct osmo_fsm_inst *fi, int out_event, struct xua_msg *in
 		/* RFC 3868 Ch. 3.5.1 */
 		xua->hdr = XUA_HDR(SUA_MSGC_ASPSM, SUA_ASPSM_UP);
 		/* Optional: ASP ID */
-		if (asp->asp_id_present)
-			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp->asp_id);
+#if 0
+		/* TODO: RFC 4666 3.8.1:
+		 * 'The "ASP Identifier Required" error is sent by an SGP in
+		 * response to an ASP Up message that does not contain an ASP
+		 * Identifier parameter when the SGP requires one. The ASP SHOULD
+		 * resend the ASP Up message with an ASP Identifier.' */
+		if (ss7_asp_peer_requires_asp_id(asp)) { /* Maybe configure in VTY "asp" node? */
+			asp_id = /* get a unique id of asp within as, eg. the index in as->asps[] */;
+			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp_id);
+		}
+#endif
 		/* Optional: Info String */
 		break;
 	case XUA_ASP_E_ASPSM_ASPUP_ACK:
 		/* RFC3868 Ch. 3.5.2 */
 		xua->hdr = XUA_HDR(SUA_MSGC_ASPSM, SUA_ASPSM_UP_ACK);
+		/* Optional: ASP ID */
+#if 0
+		/* TODO: RFC 4666 3.5.2:
+		 * "The optional ASP Identifier parameter is specifically useful for IPSP
+		 * communication.  In that case, the IPSP answering the ASP Up message
+		 * MAY include its own ASP Identifier value." */
+		if (ss7_asp_peer_requires_asp_id(asp)) { /* Maybe configure in VTY "asp" node? */
+			asp_id = /* get a unique id of asp within as, eg. the index in as->asps[] */;
+			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp_id);
+		}
+#endif
 		/* Optional: Info String */
 		break;
 	case XUA_ASP_E_ASPSM_ASPDN:
@@ -249,7 +282,7 @@ static int peer_send(struct osmo_fsm_inst *fi, int out_event, struct xua_msg *in
 		/* RFC3868 Ch. 3.6.1 */
 		xua->hdr = XUA_HDR(SUA_MSGC_ASPTM, SUA_ASPTM_ACTIVE);
 		/* Optional: Traffic Mode Type */
-		rc = determine_traf_mode(asp);
+		rc = ss7_asp_determine_traf_mode(asp);
 		if (rc >= 0)
 			xua_msg_add_u32(xua, M3UA_IEI_TRAF_MODE_TYP, osmo_ss7_tmode_to_xua(rc));
 		/* Optional: Routing Context */
@@ -315,6 +348,7 @@ static void xua_t_ack_cb(void *data)
 {
 	struct osmo_fsm_inst *fi = data;
 	struct xua_asp_fsm_priv *xafp = fi->priv;
+	uint32_t timeout_sec;
 
 	LOGPFSML(fi, LOGL_INFO, "T(ack) callback: re-transmitting event %s\n",
 		osmo_fsm_event_name(fi->fsm, xafp->t_ack.out_event));
@@ -323,13 +357,15 @@ static void xua_t_ack_cb(void *data)
 	peer_send(fi, xafp->t_ack.out_event, NULL);
 
 	/* Re-start the timer */
-	osmo_timer_schedule(&xafp->t_ack.timer, XUA_T_ACK_SEC, 0);
+	timeout_sec = osmo_tdef_get(xafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_ACK, OSMO_TDEF_S, -1);
+	osmo_timer_schedule(&xafp->t_ack.timer, timeout_sec, 0);
 }
 
 static int peer_send_and_start_t_ack(struct osmo_fsm_inst *fi,
 				     int out_event)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
+	uint32_t timeout_sec;
 	int rc;
 
 	rc = peer_send(fi, out_event, NULL);
@@ -340,7 +376,8 @@ static int peer_send_and_start_t_ack(struct osmo_fsm_inst *fi,
 	xafp->t_ack.timer.cb = xua_t_ack_cb,
 	xafp->t_ack.timer.data = fi;
 
-	osmo_timer_schedule(&xafp->t_ack.timer, XUA_T_ACK_SEC, 0);
+	timeout_sec = osmo_tdef_get(xafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_ACK, OSMO_TDEF_S, -1);
+	osmo_timer_schedule(&xafp->t_ack.timer, timeout_sec, 0);
 
 	return rc;
 }
@@ -355,7 +392,7 @@ static const uint32_t evt_ack_map[_NUM_XUA_ASP_E] = {
 
 /* Helper function to dispatch an ASP->AS event to all AS of which this
  * ASP is a memmber.  Ignores routing contexts for now. */
-static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
+static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
 	struct osmo_ss7_asp *asp = xafp->asp;
@@ -365,7 +402,7 @@ static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
 	llist_for_each_entry(as, &inst->as_list, list) {
 		if (!osmo_ss7_as_has_asp(as, asp))
 			continue;
-		osmo_fsm_inst_dispatch(as->fi, event, asp);
+		osmo_fsm_inst_dispatch(as->fi, event, data);
 	}
 }
 
@@ -385,6 +422,96 @@ static void check_stop_t_ack(struct osmo_fsm_inst *fi, uint32_t event)
 	}
 }
 
+static void xua_t_beat_stop(struct osmo_fsm_inst *fi)
+{
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+
+	LOGPFSML(fi, LOGL_DEBUG, "T(beat) stopped\n");
+	osmo_timer_del(&xafp->t_beat.timer);
+	xafp->t_beat.unacked_beats = 0;
+}
+
+static void xua_t_beat_send(struct osmo_fsm_inst *fi)
+{
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	uint32_t timeout_sec;
+
+	timeout_sec = osmo_tdef_get(xafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_BEAT, OSMO_TDEF_S, -1);
+
+	/* T(beat) disabled */
+	if (timeout_sec == 0) {
+		xua_t_beat_stop(fi);
+		return;
+	}
+
+	LOGPFSML(fi, LOGL_DEBUG, "Tx HEARTBEAT (%u unacked)\n", xafp->t_beat.unacked_beats);
+
+	/* Avoid increasing in case some extra gratuitous PING is transmitted: */
+	if (!osmo_timer_pending(&xafp->t_beat.timer))
+		xafp->t_beat.unacked_beats++;
+
+	/* (re-)arm T(beat): */
+	osmo_timer_schedule(&xafp->t_beat.timer, timeout_sec, 0);
+
+	/* Send HEARTBEAT: */
+	peer_send(fi, XUA_ASP_E_ASPSM_BEAT, NULL);
+
+}
+
+static void xua_t_beat_cb(void *_fi)
+{
+	struct osmo_fsm_inst *fi = _fi;
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+
+	if (xafp->t_beat.unacked_beats < 2) {
+		if (xafp->t_beat.unacked_beats == 1)
+			LOGPFSML(fi, LOGL_NOTICE,
+				 "Peer didn't respond to HEARTBEAT with HEARTBEAT ACK, retrying once more\n");
+		xua_t_beat_send(fi);
+		return;
+	}
+
+	/* RFC4666 4.3.4.6: If no Heartbeat Ack message (or any other M3UA
+	 * message) is received from the M3UA peer within 2*T(beat), the remote
+	 * M3UA peer is considered unavailable.  Transmission of Heartbeat
+	 * messages is stopped, and the signalling process SHOULD attempt to
+	 * re-establish communication if it is configured as the client for the
+	 * disconnected M3UA peer.
+	 */
+	LOGPFSML(fi, LOGL_NOTICE, "Peer didn't respond to HEARTBEAT with HEARTBEAT ACK and became disconnected\n");
+	ss7_asp_disconnect_stream(xafp->asp);
+
+}
+
+static void common_asp_fsm_down_onenter(struct osmo_ss7_asp *asp)
+{
+	struct osmo_ss7_as *as, *as2;
+	struct osmo_ss7_instance *inst = asp->inst;
+
+	/* First notify all AS associated to the ASP that it went down: */
+	dispatch_to_all_as(asp->fi, XUA_ASPAS_ASP_DOWN_IND, asp);
+
+	/* Implicit clean up tasks: */
+	llist_for_each_entry_safe(as, as2, &inst->as_list, list) {
+		if (!osmo_ss7_as_has_asp(as, asp))
+			continue;
+
+#ifdef WITH_TCAP_LOADSHARING
+			tcap_as_del_asp(as, asp);
+#endif
+
+		if (as->rkm_dyn_allocated) {
+			/* RFC 4666 4.4.2: "An ASP SHOULD deregister from all Application Servers
+			 * of which it is a member before attempting to move to the ASP-Down state [...]
+			 * If a Deregistration results in no more ASPs in an Application Server,
+			 * an SG MAY delete the Routing Key data."
+			 * In case it didn't deregsitrer explicitly, make sure to implicitly deregister it:
+			 */
+			ss7_as_del_asp(as, asp);
+		}
+	}
+}
+
 #define ENSURE_ASP_OR_IPSP(fi, event) 					\
 	do {								\
 		struct xua_asp_fsm_priv *_xafp = fi->priv;		\
@@ -418,7 +545,10 @@ static void check_stop_t_ack(struct osmo_fsm_inst *fi, uint32_t event)
 
 static void xua_asp_fsm_down_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_DOWN_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	xua_t_beat_stop(fi);
+	common_asp_fsm_down_onenter(asp);
 }
 
 static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -439,6 +569,11 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_ASPSM_ASPUP_ACK:
 		/* only if role ASP */
 		ENSURE_ASP_OR_IPSP(fi, event);
+		/* Optional ASP Identifier */
+		if ((asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID))) {
+			asp->remote_asp_id = xua_msg_part_get_u32(asp_id_ie);
+			asp->remote_asp_id_present = true;
+		}
 		osmo_fsm_inst_state_chg(fi, XUA_ASP_S_INACTIVE, 0, 0);
 		/* inform layer manager */
 		send_xlm_prim_simple(fi, OSMO_XLM_PRIM_M_ASP_UP, PRIM_OP_CONFIRM);
@@ -450,11 +585,10 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_ASPSM_ASPUP:
 		/* only if role SG */
 		ENSURE_SG_OR_IPSP(fi, event);
-		asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID);
 		/* Optional ASP Identifier: Store for NTFY */
-		if (asp_id_ie) {
-			asp->asp_id = xua_msg_part_get_u32(asp_id_ie);
-			asp->asp_id_present = true;
+		if ((asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID))) {
+			asp->remote_asp_id = xua_msg_part_get_u32(asp_id_ie);
+			asp->remote_asp_id_present = true;
 		}
 		/* send ACK */
 		peer_send(fi, XUA_ASP_E_ASPSM_ASPUP_ACK, NULL);
@@ -478,7 +612,27 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 
 static void xua_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	bool went_up = (prev_state == XUA_ASP_S_DOWN);
+
+	if (went_up) {
+		/* Now we are done with IPA handshake, Start Hearbeat Procedure, T(beat): */
+		xua_t_beat_send(fi);
+	}
+
+	/* RFC4666 4.3.4.5: "When an ASP moves from ASP-DOWN to ASP-INACTIVE within a
+	 * particular AS, a Notify message SHOULD be sent, by the ASP-UP receptor,
+	 * after sending the ASP-UP-ACK, in order to inform the ASP of the current AS
+	 * state."
+	 * NOTIFY is only transmitted by roles SG and IPSP.
+	 */
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = asp,
+		.asp_requires_notify = (asp->cfg.role != OSMO_SS7_ASP_ROLE_ASP) &&
+				       went_up,
+	};
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
 }
 
 static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -535,6 +689,9 @@ static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 				uint32_t rctx = osmo_load32be(&part->dat[i * sizeof(uint32_t)]);
 				as = osmo_ss7_as_find_by_rctx(asp->inst, rctx);
 				if (!as) {
+					LOGPFSML(fi, LOGL_NOTICE,
+						 "ASPAC: Couldn't find any AS with rctx=%u. Check your config!\n",
+						 rctx);
 					peer_send_error(fi, M3UA_ERR_INVAL_ROUT_CTX);
 					return;
 				}
@@ -596,7 +753,22 @@ static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 
 static void xua_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+
+	if (asp->cfg.role == OSMO_SS7_ASP_ROLE_ASP && asp->cfg.daud_act) {
+		/* RFC4666 4.6, RFC3868 4.6: "The ASP MAY choose to audit the availability
+		 * of unavailable destinations by sending DAUD messages.
+		 * This would be the case when, for example, an AS becomes active at an ASP
+		 * and does not have current destination statuses."
+		 * See also RFC4666 4.5.3, RFC3868 4.5.3 "ASP Auditing".
+		 * See also RFC4666 5.5.1.1.3 "Support for ASP Querying of SS7 Destination States"
+		 */
+		LOGPFSML(fi, LOGL_INFO, "Tx DAUD\n");
+		xua_asp_tx_snm_daud_address_book(asp);
+	}
+
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND, asp);
 }
 
 static void xua_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -693,7 +865,8 @@ static void xua_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		peer_send(fi, XUA_ASP_E_ASPSM_BEAT_ACK, xua);
 		break;
 	case XUA_ASP_E_ASPSM_BEAT_ACK:
-		/* FIXME: stop timer, if any */
+		LOGPFSML(fi, LOGL_DEBUG, "Rx HEARTBEAT ACK\n");
+		xafp->t_beat.unacked_beats = 0;
 		break;
 	case XUA_ASP_E_AS_ASSIGNED:
 		/* Ignore, only used in IPA asps so far. */
@@ -714,7 +887,14 @@ static void xua_asp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cau
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
 
+	if (!xafp)
+		return;
+
 	osmo_timer_del(&xafp->t_ack.timer);
+	xua_t_beat_stop(fi);
+
+	if (xafp->asp)
+		xafp->asp->fi = NULL;
 }
 
 static const struct osmo_fsm_state xua_asp_states[] = {
@@ -779,16 +959,16 @@ struct osmo_fsm xua_asp_fsm = {
 	.cleanup = xua_asp_fsm_cleanup,
 };
 
-static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
-					       enum osmo_ss7_asp_role role, int log_level);
+static int ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
+			     enum osmo_ss7_asp_role role, int log_level);
 
-/*! \brief Start a new ASP finite stae machine for given ASP
+/*! \brief Start a new ASP finite state machine for given ASP (stored in asp->fi)
  *  \param[in] asp Application Server Process for which to start FSM
  *  \param[in] role Role (ASP, SG, IPSP) of this FSM
  *  \param[in] log_level Logging Level for ASP FSM logging
- *  \returns FSM instance on success; NULL on error */
-struct osmo_fsm_inst *xua_asp_fsm_start(struct osmo_ss7_asp *asp,
-					enum osmo_ss7_asp_role role, int log_level)
+ *  \returns 0 on success; negative on error */
+int xua_asp_fsm_start(struct osmo_ss7_asp *asp,
+		      enum osmo_ss7_asp_role role, int log_level)
 {
 	struct osmo_fsm_inst *fi;
 	struct xua_asp_fsm_priv *xafp;
@@ -798,18 +978,25 @@ struct osmo_fsm_inst *xua_asp_fsm_start(struct osmo_ss7_asp *asp,
 
 	/* allocate as child of AS? */
 	fi = osmo_fsm_inst_alloc(&xua_asp_fsm, asp, NULL, log_level, asp->cfg.name);
+	if (!fi)
+		return -EINVAL;
 
 	xafp = talloc_zero(fi, struct xua_asp_fsm_priv);
 	if (!xafp) {
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
-		return NULL;
+		return -ENOMEM;
 	}
 	xafp->role = role;
 	xafp->asp = asp;
 
+	osmo_timer_setup(&xafp->t_beat.timer, xua_t_beat_cb, fi);
+
 	fi->priv = xafp;
 
-	return fi;
+	/* Attach FSM to ASP: */
+	asp->fi = fi;
+
+	return 0;
 }
 
 
@@ -844,8 +1031,11 @@ struct ipa_asp_fsm_priv {
 
 	/* Structure holding parsed data of the IPA CCM ID exchange */
 	struct ipaccess_unit *ipa_unit;
-	/* Timer for tracking if no PONG is received in response to PING */
-	struct osmo_timer_list pong_timer;
+	/* Timer for tracking PING without PONG response */
+	struct {
+		struct osmo_timer_list timer;
+		uint32_t unacked_beats;
+	} t_beat;
 	/* Did we receive IPA ID ACK before IPA ID RESP ? */
 	bool ipa_id_ack_rcvd;
 };
@@ -872,11 +1062,82 @@ static int get_fd_from_iafp(struct ipa_asp_fsm_priv *iafp)
 	return fd;
 }
 
+static void ipa_t_beat_stop(struct osmo_fsm_inst *fi)
+{
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+
+	LOGPFSML(fi, LOGL_DEBUG, "T(beat) stopped\n");
+	osmo_timer_del(&iafp->t_beat.timer);
+	iafp->t_beat.unacked_beats = 0;
+}
+
+static void ipa_t_beat_send(struct osmo_fsm_inst *fi)
+{
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+	struct msgb *msg;
+	uint32_t timeout_sec;
+
+	timeout_sec = osmo_tdef_get(iafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_BEAT, OSMO_TDEF_S, -1);
+
+	/* T(beat) disabled */
+	if (timeout_sec == 0) {
+		ipa_t_beat_stop(fi);
+		return;
+	}
+
+	LOGPFSML(fi, LOGL_DEBUG, "Tx HEARTBEAT (%u unacked)\n", iafp->t_beat.unacked_beats);
+
+	/* Avoid increasing in case some extra gratuitous PING is transmitted: */
+	if (!osmo_timer_pending(&iafp->t_beat.timer))
+		iafp->t_beat.unacked_beats++;
+
+	/* (re-)arm T(beat): */
+	osmo_timer_schedule(&iafp->t_beat.timer, timeout_sec, 0);
+
+	/* Send PING: */
+	if ((msg = ipa_gen_ping()))
+		osmo_ss7_asp_send(iafp->asp, msg);
+	/* we don't own msg anymore in any case here */
+}
+
+static void ipa_t_beat_cb(void *_fi)
+{
+	struct osmo_fsm_inst *fi = _fi;
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+
+	if (iafp->t_beat.unacked_beats < 2) {
+		if (iafp->t_beat.unacked_beats == 1)
+			LOGPFSML(fi, LOGL_NOTICE,
+				 "Peer didn't respond to PING with PONG, retrying once more\n");
+		ipa_t_beat_send(fi);
+		return;
+	}
+
+	/* RFC4666 4.3.4.6: If no Heartbeat Ack message (or any other M3UA
+	 * message) is received from the M3UA peer within 2*T(beat), the remote
+	 * M3UA peer is considered unavailable.  Transmission of Heartbeat
+	 * messages is stopped, and the signalling process SHOULD attempt to
+	 * re-establish communication if it is configured as the client for the
+	 * disconnected M3UA peer.
+	 */
+	LOGPFSML(fi, LOGL_NOTICE, "Peer didn't respond to PING with PONG and became disconnected\n");
+	ss7_asp_disconnect_stream(iafp->asp);
+
+}
+
 /***************
 ** FSM states **
 ***************/
 
 /* Server + Client: Initial State, wait for M-ASP-UP.req */
+static void ipa_asp_fsm_down_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	ipa_t_beat_stop(fi);
+	common_asp_fsm_down_onenter(asp);
+}
+
 static void ipa_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ipa_asp_fsm_priv *iafp = fi->priv;
@@ -931,11 +1192,15 @@ static void ipa_asp_fsm_wait_id_resp(struct osmo_fsm_inst *fi, uint32_t event, v
 		}
 		as = osmo_ss7_as_find_by_name(asp->inst, iafp->ipa_unit->unit_name);
 		if (!as) {
-			LOGPFSML(fi, LOGL_NOTICE, "Cannot find any definition for IPA Unit Name '%s'\n",
+			LOGPFSML(fi, LOGL_NOTICE, "Cannot find any AS definition for IPA Unit Name '%s'\n",
 				iafp->ipa_unit->unit_name);
 			goto out_err;
 		}
-		osmo_ss7_as_add_asp(as, asp->cfg.name);
+		rc = ss7_as_add_asp(as, asp);
+		if (rc < 0) {
+			LOGPFSML(fi, LOGL_ERROR, "Cannot add ASP '%s' to AS '%s'\n", asp->cfg.name, as->cfg.name);
+			goto out_err;
+		}
 		/* TODO: OAP Authentication? */
 		/* Send ID_ACK */
 		if (fd >= 0) {
@@ -1038,8 +1303,18 @@ static void ipa_asp_fsm_wait_id_ack(struct osmo_fsm_inst *fi, uint32_t event, vo
 
 static void ipa_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+	struct osmo_ss7_asp *asp = iafp->asp;
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = asp,
+		.asp_requires_notify = false,
+	};
+
+	/* Now we are done with IPA handshake, Start Hearbeat Procedure, T(beat): */
+	ipa_t_beat_send(fi);
+
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND, asp);
 }
 
 /* Server + Client: We're actively transmitting user data */
@@ -1055,7 +1330,12 @@ static void ipa_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *d
 
 static void ipa_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = iafp->asp,
+		.asp_requires_notify = false,
+	};
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
 }
 
 static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -1067,6 +1347,36 @@ static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 	}
 }
 
+/* Assign a 4 bit SLS (as unqiue as possible) for incoming IPA PDUs.*/
+static void _ipa_asp_pick_unused_sls(struct osmo_ss7_asp *asp, const struct osmo_ss7_as *as)
+{
+	for (unsigned int sls = 0; sls <= 0x0f; sls++) {
+		bool used = false;
+		for (unsigned i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
+			if (!as->cfg.asps[i])
+				continue;
+			if (as->cfg.asps[i] == asp)
+				continue;
+			if (!as->cfg.asps[i]->ipa.sls_assigned)
+				continue;
+			if (as->cfg.asps[i]->ipa.sls == sls) {
+				used = true;
+				break;
+			}
+		}
+		if (used)
+			continue;
+		/* Found an unused SLS, use it: */
+		asp->ipa.sls = sls;
+		asp->ipa.sls_assigned = true;
+		LOGPASP(asp, DLSS7, LOGL_DEBUG, "Assigned unsued SLS = %u\n", sls);
+		return;
+	}
+	LOGPASP(asp, DLSS7, LOGL_INFO, "All SLS in IPA AS picked, unique SLS not possible, picking random one\n");
+	asp->ipa.sls = rand() & 0x0f;
+	asp->ipa.sls_assigned = true;
+}
+
 static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ipa_asp_fsm_priv *iafp = fi->priv;
@@ -1087,12 +1397,15 @@ static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 			ipaccess_send_pong(fd);
 		break;
 	case XUA_ASP_E_ASPSM_BEAT_ACK:
-		/* stop timer, if any */
-		osmo_timer_del(&iafp->pong_timer);
+		LOGPFSML(fi, LOGL_DEBUG, "Rx HEARTBEAT ACK\n");
+		iafp->t_beat.unacked_beats = 0;
 		break;
 	case XUA_ASP_E_AS_ASSIGNED:
 		as = data;
 		osmo_talloc_replace_string(iafp->ipa_unit, &iafp->ipa_unit->unit_name, as->cfg.name);
+		/* Now that AS is known, try picking an unused SLS inside the AS.
+		 * It will be applied to PDUs received from the IPA socket. */
+		_ipa_asp_pick_unused_sls(iafp->asp, as);
 		/* Now that the AS is known, start the client side: */
 		if (iafp->role == OSMO_SS7_ASP_ROLE_ASP && fi->state == IPA_ASP_S_DOWN) {
 			LOGPFSML(fi, LOGL_NOTICE, "Bringing up ASP now once it has been assigned to an AS\n");
@@ -1104,16 +1417,6 @@ static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	}
 }
 
-static void ipa_pong_timer_cb(void *_fi)
-{
-	struct osmo_fsm_inst *fi = _fi;
-	struct ipa_asp_fsm_priv *iafp = fi->priv;
-
-	LOGPFSML(fi, LOGL_NOTICE, "Peer didn't respond to PING? with PONG!\n");
-	/* kill ASP and (wait for) re-connect */
-	osmo_ss7_asp_disconnect(iafp->asp);
-}
-
 static int ipa_asp_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct ipa_asp_fsm_priv *iafp = fi->priv;
@@ -1124,6 +1427,19 @@ static int ipa_asp_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	return -1;
 }
 
+static void ipa_asp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+{
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+
+	if (!iafp)
+		return;
+
+	ipa_t_beat_stop(fi);
+
+	if (iafp->asp)
+		iafp->asp->fi = NULL;
+}
+
 static const struct osmo_fsm_state ipa_asp_states[] = {
 	[IPA_ASP_S_DOWN] = {
 		.in_event_mask = S(XUA_ASP_E_M_ASP_UP_REQ) |
@@ -1132,7 +1448,7 @@ static const struct osmo_fsm_state ipa_asp_states[] = {
 				  S(IPA_ASP_S_WAIT_ID_RESP),
 		.name = "ASP_DOWN",
 		.action = ipa_asp_fsm_down,
-		.onenter = xua_asp_fsm_down_onenter,
+		.onenter = ipa_asp_fsm_down_onenter,
 	},
 	/* Server Side */
 	[IPA_ASP_S_WAIT_ID_RESP] = {
@@ -1200,16 +1516,17 @@ struct osmo_fsm ipa_asp_fsm = {
 			       S(XUA_ASP_E_ASPSM_BEAT_ACK) |
 			       S(XUA_ASP_E_AS_ASSIGNED),
 	.allstate_action = ipa_asp_allstate,
+	.cleanup = ipa_asp_fsm_cleanup,
 };
 
 
-/*! \brief Start a new ASP finite state machine for given ASP
+/*! \brief Start a new ASP finite state machine for given ASP (stored on asp->fi)
  *  \param[in] asp Application Server Process for which to start FSM
  *  \param[in] role Role (ASP, SG, IPSP) of this FSM
  *  \param[in] log_level Logging Level for ASP FSM logging
- *  \returns FSM instance on success; NULL on error */
-static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
-					       enum osmo_ss7_asp_role role, int log_level)
+ *  \returns 0 on success; negative on error */
+static int ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
+			     enum osmo_ss7_asp_role role, int log_level)
 {
 	struct osmo_fsm_inst *fi;
 	struct ipa_asp_fsm_priv *iafp;
@@ -1219,19 +1536,24 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 
 	/* allocate as child of AS? */
 	fi = osmo_fsm_inst_alloc(&ipa_asp_fsm, asp, NULL, log_level, asp->cfg.name);
+	if (!fi)
+		return -EINVAL;
 
 	iafp = talloc_zero(fi, struct ipa_asp_fsm_priv);
 	if (!iafp) {
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
-		return NULL;
+		return -ENOMEM;
 	}
 
 	if (as) {
 		unit_name = as->cfg.name;
+		/* Allocacate potentially unique SLS within AS since AS is already known: */
+		_ipa_asp_pick_unused_sls(asp, as);
 	} else if (asp->dyn_allocated) {
 		LOGPFSML(fi, LOGL_INFO, "Dynamic ASP is not assigned to any AS, "
 			 "using ASP name instead of AS name as ipa_unit_name\n");
 		unit_name = asp->cfg.name;
+		/* asp->ipa.sls will be assigned together with AS unit_name during XUA_ASP_E_AS_ASSIGNED. */
 	} else {
 		/* ASP in client mode will be brought up when this ASP is added
 		 * to an AS, see XUA_ASP_E_AS_ASSIGNED. */
@@ -1240,19 +1562,22 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 			can_start = false;
 		}
 		unit_name = asp->cfg.name;
+		/* asp->ipa.sls will be assigned together with AS unit_name during XUA_ASP_E_AS_ASSIGNED. */
 	}
 
 	iafp->role = role;
 	iafp->asp = asp;
 	iafp->ipa_unit = talloc_zero(iafp, struct ipaccess_unit);
 	iafp->ipa_unit->unit_name = talloc_strdup(iafp->ipa_unit, unit_name);
-	iafp->pong_timer.cb = ipa_pong_timer_cb;
-	iafp->pong_timer.data = fi;
+	osmo_timer_setup(&iafp->t_beat.timer, ipa_t_beat_cb, fi);
 
 	fi->priv = iafp;
 
+	/* Attach FSM to ASP: */
+	asp->fi = fi;
+
 	if (can_start && role == OSMO_SS7_ASP_ROLE_ASP)
 		osmo_fsm_inst_dispatch(fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
 
-	return fi;
+	return 0;
 }
