@@ -37,7 +37,8 @@ void sua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned in
 void sua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 		 uint32_t dpc, uint16_t user, uint16_t cause, const char *info_str);
 
-int m3ua_hmdc_rx_from_l2(struct osmo_ss7_instance *inst, struct xua_msg *xua);
+struct xua_msg *sua_gen_cldr(const struct xua_msg *xua_in, uint32_t route_ctx, uint32_t ret_cause);
+
 struct msgb *m3ua_to_msg(struct xua_msg *xua);
 int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua);
 void m3ua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
@@ -52,6 +53,10 @@ void m3ua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned i
 void m3ua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 		  uint32_t dpc, uint16_t user, uint16_t cause, const char *info_str);
 
+struct xua_msg *m3ua_encode_duna(const uint32_t *rctx, unsigned int num_rctx,
+				 const uint32_t *aff_pc, unsigned int num_aff_pc,
+				 const char *info_string);
+
 void xua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 			  const uint32_t *aff_pc, unsigned int num_aff_pc,
 			  const char *info_str, bool available);
@@ -71,9 +76,6 @@ void m3ua_dh_to_xfer_param(struct osmo_mtp_transfer_param *param,
 			   const struct m3ua_data_hdr *mdh);
 void mtp_xfer_param_to_m3ua_dh(struct m3ua_data_hdr *mdh,
 				const struct osmo_mtp_transfer_param *param);
-int hmrt_mtp_xfer_request_l4_to_l3(struct osmo_ss7_instance *inst,
-				   const struct osmo_mtp_transfer_param *param,
-				   uint8_t *user_data, size_t user_data_len);
 
 
 extern const struct xua_msg_class m3ua_msg_class_mgmt;
@@ -92,7 +94,6 @@ struct xua_msg *m3ua_encode_notify(const struct osmo_xlm_prim_notify *npar);
 int m3ua_decode_notify(struct osmo_xlm_prim_notify *npar, void *ctx,
 			const struct xua_msg *xua);
 int m3ua_rx_rkm(struct osmo_ss7_asp *asp, struct xua_msg *xua);
-void xua_rkm_cleanup_dyn_as_for_asp(struct osmo_ss7_asp *asp);
 
 struct osmo_xlm_prim *xua_xlm_prim_alloc(enum osmo_xlm_prim_type prim_type,
 					 enum osmo_prim_operation op);
