@@ -151,6 +151,19 @@ static void xua_rkm_send_dereg_req(struct osmo_ss7_asp *asp, uint32_t route_ctx)
  * RKM REG request */
 #define MAX_NEW_AS 16
 
+/* Lookup "as" pointer in array "newly_assigned_as" with "num_newly_assigned_as" elements. */
+static bool newly_assigned_as_in_array(struct osmo_ss7_as * const *newly_assigned_as,
+				       unsigned int num_newly_assigned_as,
+				       const struct osmo_ss7_as *as)
+{
+	unsigned int i;
+	for (i = 0; i < num_newly_assigned_as; i++) {
+		if (as == newly_assigned_as[i])
+			return true;
+	}
+	return false;
+}
+
 /* SG: handle a single registration request IE (nested IEs in 'innner' */
 static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 			   struct msgb *resp, struct osmo_ss7_as **newly_assigned_as,
@@ -161,6 +174,7 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 	struct osmo_ss7_as *as;
 	struct osmo_ss7_route *rt;
 	char namebuf[32];
+	bool as_already_in_array;
 
 	/* mandatory local routing key ID */
 	rk_id = xua_msg_get_u32(inner, M3UA_IEI_LOC_RKEY_ID);
@@ -219,15 +233,27 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 
 	/* check if there is already an AS for this routing key */
 	as = osmo_ss7_as_find_by_rctx(asp->inst, rctx);
+
+	if (!as && !asp->inst->cfg.permit_dyn_rkm_alloc) {
+		/* not permitted to create dynamic RKM entries */
+		LOGPASP(asp, DLSS7, LOGL_NOTICE, "RKM: RCTX %u not found in configuration, and "
+			"dynamic RKM allocation not permitted; permission denied\n", rctx);
+		msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_PERM_DENIED, 0);
+		return -1;
+	}
+
+	/* Early return before allocating stuff if no space left.
+	 * If AS didn't exist before, it can't be already in the array:*/
+	as_already_in_array = as && newly_assigned_as_in_array(newly_assigned_as, *nas_idx, as);
+	if (!as_already_in_array && *nas_idx >= max_nas_idx) {
+		LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
+			max_nas_idx+1);
+		msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
+		return -1;
+	}
+
 	if (as) {
 		LOGPASP(asp, DLSS7, LOGL_NOTICE, "RKM: Found existing AS for RCTX %u\n", rctx);
-		/* Early return before allocating stuff if no space left: */
-		if (*nas_idx >= max_nas_idx) {
-			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
-				max_nas_idx+1);
-			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
-			return -1;
-		}
 
 		if (as->cfg.routing_key.pc != dpc) {
 			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: DPC doesn't match, rejecting AS (%u != %u)\n",
@@ -251,15 +277,7 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 			}
 			as->cfg.mode_set_by_peer = true;
 		}
-	} else if (asp->inst->cfg.permit_dyn_rkm_alloc) {
-		/* Early return before allocating stuff if no space left: */
-		if (*nas_idx >= max_nas_idx) {
-			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: not enough room for newly assigned AS (max %u AS)\n",
-				max_nas_idx+1);
-			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
-			return -1;
-		}
-
+	} else {
 		/* Create an AS for this routing key */
 		snprintf(namebuf, sizeof(namebuf), "as-rkm-%u", rctx);
 		as = osmo_ss7_as_find_or_create(asp->inst, namebuf, OSMO_SS7_ASP_PROT_M3UA);
@@ -279,8 +297,8 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 		as->cfg.routing_key.pc = dpc;
 		as->cfg.routing_key.context = rctx;
 
-		/* add route for that routing key */
-		rt = ss7_route_create(as->inst->rtable_system, dpc, 0xFFFFFF, namebuf);
+		/* add dynamic route for that routing key */
+		rt = ss7_route_create(as->inst->rtable_system, dpc, 0xFFFFFF, true, namebuf);
 		if (!rt) {
 			LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: Cannot insert route for DPC %s / as %s\n",
 				osmo_ss7_pointcode_print(asp->inst, dpc), namebuf);
@@ -288,19 +306,19 @@ static int handle_rkey_reg(struct osmo_ss7_asp *asp, struct xua_msg *inner,
 			msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_CANT_SUPP_UNQ_RT, 0);
 			return -1;
 		}
-	} else {
-		/* not permitted to create dynamic RKM entries */
-		LOGPASP(asp, DLSS7, LOGL_NOTICE, "RKM: RCTX %u not found in configuration, and "
-			"dynamic RKM allocation not permitted; permission denied\n", rctx);
-		msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_PERM_DENIED, 0);
-		return -1;
 	}
 
 	/* Success: Add just-create AS to connected ASP + report success */
-	ss7_as_add_asp(as, asp);
+	if (ss7_as_add_asp(as, asp) < 0) {
+		LOGPASP(asp, DLSS7, LOGL_ERROR, "RKM: Cannot associate ASP to AS %s\n", as->cfg.name);
+		msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_ERR_INSUFF_RESRC, 0);
+		return -1;
+	}
+
 	msgb_append_reg_res(resp, rk_id, M3UA_RKM_REG_SUCCESS, rctx);
 	/* append to list of newly assigned as */
-	newly_assigned_as[(*nas_idx)++] = as;
+	if (!as_already_in_array)
+		newly_assigned_as[(*nas_idx)++] = as;
 	return 0;
 }
 
@@ -310,7 +328,7 @@ static int m3ua_rx_rkm_reg_req(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	struct xua_msg_part *part;
 	struct msgb *resp = m3ua_msgb_alloc(__func__);
 	struct osmo_ss7_as *newly_assigned_as[MAX_NEW_AS];
-	unsigned int i, nas_idx = 0;
+	unsigned int i, num_newly_assigned_as = 0;
 
 	memset(newly_assigned_as, 0, sizeof(newly_assigned_as));
 
@@ -330,7 +348,7 @@ static int m3ua_rx_rkm_reg_req(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 		/* handle single registration and append result to
 		 * 'resp' */
 		handle_rkey_reg(asp, inner, resp, newly_assigned_as,
-				ARRAY_SIZE(newly_assigned_as), &nas_idx);
+				ARRAY_SIZE(newly_assigned_as), &num_newly_assigned_as);
 
 		xua_msg_free(inner);
 	}
@@ -341,7 +359,7 @@ static int m3ua_rx_rkm_reg_req(struct osmo_ss7_asp *asp, struct xua_msg *xua)
 	/* and *after* the RKM REG Response inform the newly assigned
 	 * ASs about the fact that there's an INACTIVE ASP for them,
 	 * which will cause them to send NOTIFY to the client */
-	for (i = 0; i < ARRAY_SIZE(newly_assigned_as); i++) {
+	for (i = 0; i < num_newly_assigned_as; i++) {
 		struct osmo_ss7_as *as = newly_assigned_as[i];
 		if (!as)
 			continue;
@@ -393,7 +411,9 @@ static int handle_rkey_dereg(struct osmo_ss7_asp *asp, uint32_t rctx,
 		return -1;
 	}
 
-	rt = ss7_route_table_find_route_by_dpc_mask(inst->rtable_system, as->cfg.routing_key.pc, 0xffffff);
+	rt = ss7_route_table_find_route_by_dpc_mask(inst->rtable_system,
+						    as->cfg.routing_key.pc, 0xffffff,
+						    true);
 	if (!rt) {
 		msgb_append_dereg_res(resp, M3UA_RKM_DEREG_ERR_UNKNOWN, 0);
 		return -1;
@@ -402,19 +422,12 @@ static int handle_rkey_dereg(struct osmo_ss7_asp *asp, uint32_t rctx,
 	LOGPASP(asp, DLSS7, LOGL_INFO, "RKM: De-Registering rctx %u for DPC %s\n",
 		rctx, osmo_ss7_pointcode_print(inst, as->cfg.routing_key.pc));
 
-	/* remove ASP from AS */
-	osmo_ss7_as_del_asp(as, asp->cfg.name);
-	/* FIXME: Rather than spoofing teh ASP-DOWN.ind to the AS here,
-	 * we should refuse RKM DEREG if the ASP is still ACTIVE */
-	osmo_fsm_inst_dispatch(as->fi, XUA_ASPAS_ASP_DOWN_IND, asp);
+	/* Release the associated route */
+	ss7_route_destroy(rt);
+	/* Dissassociate the ASP from the dynamically allocated AS.
+	 * The AS may be freed if it is serving no more ASPs. */
+	ss7_as_del_asp(as, asp);
 
-	/* if we were dynamically allocated, release the associated
-	 * route and destroy the AS */
-	if (as->rkm_dyn_allocated) {
-		/* remove route + AS definition */
-		ss7_route_destroy(rt);
-		osmo_ss7_as_destroy(as);
-	}
 	/* report success */
 	msgb_append_dereg_res(resp, M3UA_RKM_DEREG_SUCCESS, rctx);
 
@@ -610,21 +623,3 @@ int osmo_xlm_sap_down(struct osmo_ss7_asp *asp, struct osmo_prim_hdr *oph)
 	msgb_free(prim->oph.msg);
 	return 0;
 }
-
-/* clean-up any dynamically created ASs + routes */
-void xua_rkm_cleanup_dyn_as_for_asp(struct osmo_ss7_asp *asp)
-{
-	struct osmo_ss7_instance *inst = asp->inst;
-	struct osmo_ss7_as *as, *as2;
-
-	llist_for_each_entry_safe(as, as2, &inst->as_list, list) {
-		if (!osmo_ss7_as_has_asp(as, asp))
-			continue;
-		if (!as->rkm_dyn_allocated)
-			continue;
-
-		/* If there are no other ASPs, destroy the AS: */
-		if (osmo_ss7_as_count_asp(as) == 1)
-			osmo_ss7_as_destroy(as);
-	}
-}
